from mim.utils import read_installation_records


def get_installed_package(ctx, args, incomplete):
    pkgs = []
    for pkg, _, _ in read_installation_records():
        pkgs.append(pkg)
    return pkgs


def get_downstream_package(ctx, args, incomplete):
    pkgs = []
    for pkg, _, _ in read_installation_records():
        if pkg == 'mmcv' or pkg == 'mmcv-full':
            continue
        pkgs.append(pkg)
    return pkgs


def get_official_package(ctx=None, args=None, incomplete=None):
    return [
        'mmcls',
        'mmdet',
        'mmdet3d',
        'mmseg',
        'mmaction',
        'mmtrack',
        'mmpose',
        'mmedit',
        'mmocr',
        'mmgen',
    ]



import pkg_resources
from typing import List

import click
from tabulate import tabulate

from mim.utils import read_installation_records


@click.command('list')
@click.option(
    '--all',
    is_flag=True,
    help='List packages of OpenMMLab projects or all the packages in the '
    'python environment.')
def cli(all: bool = True) -> None:
    """List packages.

    \b
    Example:
        > mim list
        > mim list --all
    """
    table_header = ['Package', 'Version', 'Source']
    table_data = list_package(all=all)
    click.echo(tabulate(table_data, headers=table_header, tablefmt='simple'))


def list_package(all: bool = False) -> List[List[str]]:
    """List packages.

    List packages of OpenMMLab projects or all the packages in the python
    environment.

    Args:
        all (bool): List all installed packages. If all is False, it just lists
            the packages installed by mim. Default: False.
    """
    if not all:
        pkgs_info = read_installation_records()
    else:
        pkgs_info = []
        for pkg in pkg_resources.working_set:
            pkgs_info.append([pkg.project_name, pkg.version])

    return pkgs_info



import copy as cp
import itertools
import os
import os.path as osp
import subprocess
import time
from concurrent.futures import ProcessPoolExecutor as Executor
from typing import Optional, Tuple, Union

import click

from mim.click import CustomCommand
from mim.utils import (
    args2string,
    echo_error,
    echo_success,
    exit_with_error,
    get_config,
    get_installed_path,
    highlighted_error,
    is_installed,
    recursively_find,
    set_config,
    string2args,
)


@click.command(
    name='gridsearch',
    context_settings=dict(ignore_unknown_options=True),
    cls=CustomCommand)
@click.argument('package', type=str)
@click.argument('config', type=str)
@click.option(
    '--launcher',
    type=click.Choice(['none', 'pytorch', 'slurm'], case_sensitive=False),
    default='none',
    help='Job launcher')
@click.option(
    '--port',
    type=int,
    default=29500,
    help=('The port used for inter-process communication '
          '(only applicable to slurm / pytorch launchers)'))
@click.option('--gpus', type=int, default=1, help='Number of gpus to use')
@click.option(
    '--gpus-per-node',
    type=int,
    help=('Number of gpus per node to use '
          '(only applicable to launcher == "slurm")'))
@click.option(
    '--cpus-per-task',
    type=int,
    default=2,
    help='Number of cpus per task (only applicable to launcher == "slurm")')
@click.option(
    '--partition',
    type=str,
    help='The partition to use (only applicable to launcher == "slurm")')
@click.option('--max-jobs', type=int, help='Max parallel number', default=1)
@click.option(
    '--srun-args', type=str, help='Other srun arguments that might be used')
@click.option(
    '--search-args', type=str, help='Arguments for hyper parameters search')
@click.argument('other_args', nargs=-1, type=click.UNPROCESSED)
def cli(package: str,
        config: str,
        gpus: int,
        gpus_per_node: int,
        partition: str,
        cpus_per_task: int = 2,
        max_jobs: int = 1,
        launcher: str = 'none',
        port: int = 29500,
        srun_args: Optional[str] = None,
        search_args: str = '',
        other_args: tuple = ()) -> None:
    """Perform Hyper-parameter search.

    Example:

    \b
    # Parameter search with on a single server with one GPU, search learning
    # rate
    > mim gridsearch mmcls resnet101_b16x8_cifar10.py --work-dir tmp --gpus \
        1 --search-args '--optimizer.lr 1e-2 1e-3'
    # Parameter search with on a single server with one GPU, search
    # weight_decay
    > mim gridsearch mmcls resnet101_b16x8_cifar10.py --work-dir tmp --gpus \
        1 --search-args '--optimizer.weight_decay 1e-3 1e-4'
    # Parameter search with on a single server with one GPU, search learning
    # rate and weight_decay
    > mim gridsearch mmcls resnet101_b16x8_cifar10.py --work-dir tmp --gpus \
        1 --search-args '--optimizer.lr 1e-2 1e-3 --optimizer.weight_decay \
        1e-3 1e-4'
    # Parameter search on a slurm HPC with one 8-GPU node, search learning
    # rate and weight_decay
    > mim gridsearch mmcls resnet101_b16x8_cifar10.py --work-dir tmp --gpus \
        8 --partition partition_name --gpus-per-node 8 --launcher slurm \
        --search-args '--optimizer.lr 1e-2 1e-3 --optimizer.weight_decay 1e-3 \
        1e-4'
    # Parameter search on a slurm HPC with one 8-GPU node, search learning
    # rate and weight_decay, max parallel jobs is 2
    > mim gridsearch mmcls resnet101_b16x8_cifar10.py --work-dir tmp --gpus \
        8 --partition partition_name --gpus-per-node 8 --launcher slurm \
        --max-jobs 2 --search-args '--optimizer.lr 1e-2 1e-3 \
        --optimizer.weight_decay 1e-3 1e-4'
    # Print the help message of sub-command search
    > mim gridsearch -h
    # Print the help message of sub-command search and the help message of the
    # training script of codebase mmcls
    > mim gridsearch mmcls -h
    """

    is_success, msg = gridsearch(
        package=package,
        config=config,
        gpus=gpus,
        gpus_per_node=gpus_per_node,
        cpus_per_task=cpus_per_task,
        max_jobs=max_jobs,
        partition=partition,
        launcher=launcher,
        port=port,
        srun_args=srun_args,
        search_args=search_args,
        other_args=other_args)

    if is_success:
        echo_success(msg)  # type: ignore
    else:
        exit_with_error(msg)


def gridsearch(
    package: str,
    config: str,
    gpus: int,
    gpus_per_node: int = None,
    cpus_per_task: int = 2,
    max_jobs: int = 1,
    partition: str = None,
    launcher: str = 'none',
    port: int = 29500,
    srun_args: Optional[str] = None,
    search_args: str = '',
    other_args: tuple = ()
) -> Tuple[bool, Union[str, Exception]]:
    """Hyper parameter search with given config.

    Args:
        package (str): The codebase name.
        config (str): The config file path. If not exists, will search in the
            config files of the codebase.
        gpus (int): Number of gpus used for training.
        gpus_per_node (int, optional): Number of gpus per node to use
            (only applicable to launcher == "slurm"). Defaults to None.
        cpus_per_task (int, optional): Number of cpus per task to use
            (only applicable to launcher == "slurm"). Defaults to None.
        partition (str, optional): The partition name
            (only applicable to launcher == "slurm"). Defaults to None.
        max_jobs (int, optional): The max number of workers. Applicable only
            if launcher == 'slurm'. Default to 1.
        launcher (str, optional): The launcher used to launch jobs.
            Defaults to 'none'.
        port (int, optional): The port used for inter-process communication
            (only applicable to slurm / pytorch launchers). Default to 29500.
        srun_args (str, optional): Other srun arguments that might be
            used, all arguments should be in a string. Defaults to None.
        search_args (str, optional): Arguments for hyper parameters search, all
            arguments should be in a string. Defaults to None.
        other_args (tuple, optional): Other arguments, will be passed to the
            codebase's training script. Defaults to ().
    """
    # If launcher == "slurm", must have following args
    if launcher == 'slurm':
        msg = ('If launcher is slurm, '
               'gpus-per-node and partition should not be None')
        flag = (gpus_per_node is not None) and (partition is not None)
        if not flag:
            raise AssertionError(highlighted_error(msg))

    if not is_installed(package):
        msg = (f'The codebase {package} is not installed, '
               'do you want to install it? ')
        if click.confirm(msg):
            click.echo(f'Installing {package}')
            cmd = ['mim', 'install', package]
            ret = subprocess.check_call(cmd)
            if ret != 0:
                msg = f'{package} is not successfully installed'
                raise RuntimeError(highlighted_error(msg))
            else:
                click.echo(f'{package} is successfully installed')
        else:
            msg = f'You can not train this model without {package} installed.'
            return False, msg

    pkg_root = get_installed_path(package)

    if not osp.exists(config):
        files = recursively_find(pkg_root, osp.basename(config))

        if len(files) == 0:
            msg = (f"The path {config} doesn't exist and we can not "
                   f'find the config file in codebase {package}.')
            raise ValueError(highlighted_error(msg))
        elif len(files) > 1:
            msg = (
                f"The path {config} doesn't exist and we find multiple "
                f'config files with same name in codebase {package}: {files}.')
            raise ValueError(highlighted_error(msg))
        click.echo(
            f"The path {config} doesn't exist but we find the config file "
            f'in codebase {package}, will use {files[0]} instead.')
        config = files[0]

    train_script = osp.join(pkg_root, 'tools/train.py')

    # parsing search_args
    # the search_args looks like:
    # "--optimizer.lr 0.001 0.01 0.1 --optimizer.weight_decay 1e-4 1e-3 1e-2"
    search_args_dict = string2args(search_args)
    if not len(search_args_dict):
        msg = 'Should specify at least one arg for searching'
        raise ValueError(highlighted_error(msg))

    for k in search_args_dict:
        if search_args_dict[k] is bool:
            msg = f'Should specify at least one value for arg {k}'
            raise ValueError(highlighted_error(msg))

    try:
        from mmcv import Config
    except ImportError:
        msg = 'Please install mmcv to use the gridsearch command.'
        raise ImportError(highlighted_error(msg))

    cfg = Config.fromfile(config)
    for arg in search_args_dict:
        try:
            arg_type = type(get_config(cfg, arg))
            search_args_dict[arg] = [eval(x) for x in search_args_dict[arg]]
            for val in search_args_dict[arg]:
                assert type(val) == arg_type
        except AssertionError:
            msg = f'Arg {arg} not in the config file. '
            raise AssertionError(highlighted_error(msg))

    other_args_dict = string2args(' '.join(other_args))

    work_dir = other_args_dict.get('work-dir')

    if work_dir:
        work_dir = work_dir[0]
    else:
        work_dir = cfg.get('work_dir')

    if work_dir is None:
        msg = 'work_dir is not specified'
        raise AssertionError(highlighted_error(msg))

    cfg.pop('work_dir', None)

    # remove redundant '/' at the end of work_dir

    assert work_dir  # To pass mypy test
    while work_dir.endswith('/'):
        work_dir = work_dir[:-1]

    config_tmpl, config_suffix = osp.splitext(osp.basename(config))
    work_dir_tmpl = work_dir

    cmds = []
    exp_names = []

    arg_names = [k for k in search_args_dict]
    arg_values = [search_args_dict[k] for k in arg_names]

    for combination in itertools.product(*arg_values):
        cur_cfg = Config(cp.deepcopy(cfg))
        suffix_list = []

        for k, v in zip(arg_names, combination):
            suffix_list.extend([k, str(v)])
            set_config(cur_cfg, k, v)

        name_suffix = '_search_' + '_'.join(suffix_list)
        work_dir = work_dir_tmpl + name_suffix
        os.makedirs(work_dir, exist_ok=True)

        config_name = config_tmpl + name_suffix + config_suffix
        exp_names.append(config_tmpl + name_suffix)
        config_path = osp.join(work_dir, config_name)

        with open(config_path, 'w') as fout:
            fout.write(cur_cfg.pretty_text)

        other_args_dict_ = cp.deepcopy(other_args_dict)
        other_args_dict_['work-dir'] = [work_dir]

        other_args_str = args2string(other_args_dict_)

        common_args = ['--launcher', launcher] + other_args_str.split()

        if launcher == 'none':
            cmd = ['python', train_script, config_path, '--gpus',
                   str(gpus)] + common_args
        elif launcher == 'pytorch':
            cmd = [
                'python', '-m', 'torch.distributed.launch',
                f'--nproc_per_node={gpus}', f'--master_port={port}',
                train_script, config_path
            ] + common_args
        elif launcher == 'slurm':
            parsed_srun_args = srun_args.split() if srun_args else []
            cmd = [
                'srun', '-p', f'{partition}', f'--gres=gpu:{gpus_per_node}',
                f'--ntasks={gpus}', f'--ntasks-per-node={gpus_per_node}',
                f'--cpus-per-task={cpus_per_task}', '--kill-on-bad-exit=1'
            ] + parsed_srun_args + ['python', '-u', train_script, config_path
                                    ] + common_args

        cmds.append(cmd)

    time.sleep(5)
    succeed_list, fail_list = [], []
    if launcher in ['none', 'pytorch']:
        for cmd, exp_name in zip(cmds, exp_names):
            cmd_text = ' '.join(cmd)
            click.echo(f'Training command for exp {exp_name} is {cmd_text}. ')

            ret = subprocess.check_call(
                cmd, env=dict(os.environ, MASTER_PORT=str(port)))
            if ret == 0:
                click.echo(f'Exp {exp_name} finished successfully.')
                succeed_list.append(exp_name)
            else:
                echo_error('Training not finished successfully.')
                fail_list.append(exp_name)

    elif launcher == 'slurm':

        with Executor(max_workers=max_jobs) as executor:

            for exp, ret in zip(exp_names,
                                executor.map(subprocess.check_call, cmds)):
                if ret == 0:
                    click.echo(f'Exp {exp} finished successfully.')
                    succeed_list.append(exp)
                else:
                    echo_error(f'Exp {exp} not finished successfully.')
                    fail_list.append(exp)

    if len(fail_list):
        msg = ('The following experiments in hyper parameter search '
               f'failed: \n {fail_list}')
        return False, msg
    else:
        msg = ('The hyper parameter search finished successfully.'
               f'Experiment list: \n {succeed_list}')
        return True, msg



import sys

import click

from mim.click import get_installed_package, param2lowercase
from mim.utils import PKG_ALIAS, call_command, remove_installation_records


@click.command('uninstall')
@click.argument(
    'package', autocompletion=get_installed_package, callback=param2lowercase)
@click.option(
    '-y',
    '--yes',
    'confirm_yes',
    is_flag=True,
    help='Don’t ask for confirmation of uninstall deletions.')
def cli(package: str, confirm_yes: bool) -> None:
    """Uninstall package.

    Example:

        > mim uninstall mmcv-full
    """
    uninstall(package, confirm_yes)


def uninstall(package: str, confirm_yes=False) -> None:
    """Uninstall package.

    Args:
        package (str): The name of uninstalled package, such as mmcv-full.
        confirm_yes (bool): Don’t ask for confirmation of uninstall deletions.
            Default: True.
    """
    uninstall_cmd = ['python', '-m', 'pip', 'uninstall', package]
    if confirm_yes:
        uninstall_cmd.append('-y')

    call_command(uninstall_cmd)
    # if package is installed, importlib.import_module will import the
    # package and add it to sys.modules. However, if we uninstall the
    # package in the same process, is_installed will give a wrong result
    # because importlib.import_module will search package from sys.modules
    # first.
    sys.modules.pop(PKG_ALIAS.get(package, package), None)

    remove_installation_records(package)



import os
import os.path as osp
import subprocess
import tempfile
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from distutils.version import LooseVersion
from pkg_resources import parse_requirements
from typing import List

import click

from mim.click import get_official_package, param2lowercase
from mim.commands.uninstall import uninstall
from mim.utils import (
    DEFAULT_URL,
    PKG2PROJECT,
    PKG_ALIAS,
    WHEEL_URL,
    call_command,
    echo_success,
    echo_warning,
    get_installed_version,
    get_latest_version,
    get_package_version,
    get_release_version,
    get_torch_cuda_version,
    highlighted_error,
    is_installed,
    is_version_equal,
    parse_url,
    split_package_version,
    write_installation_records,
)


@click.command('install')
@click.argument(
    'package',
    type=str,
    autocompletion=get_official_package,
    callback=param2lowercase)
@click.option(
    '-f', '--find', 'find_url', type=str, help='Url for finding package.')
@click.option(
    '--default-timeout',
    'timeout',
    type=int,
    default=45,
    help='Set the socket timeout (default 15 seconds).')
@click.option(
    '-y',
    '--yes',
    'is_yes',
    is_flag=True,
    help='Don’t ask for confirmation of uninstall deletions.')
@click.option(
    '--user',
    'is_user_dir',
    is_flag=True,
    help='Install to the Python user install directory')
def cli(
    package: str,
    find_url: str = '',
    timeout: int = 30,
    is_yes: bool = False,
    is_user_dir: bool = False,
) -> None:
    """Install package.

    Example:

    \b
    # install latest version of mmcv-full
    > mim install mmcv-full  # wheel
    # install 1.3.1
    > mim install mmcv-full==1.3.1
    # install master branch
    > mim install mmcv-full -f https://github.com/open-mmlab/mmcv.git

    # install latest version of mmcls
    > mim install mmcls
    # install 0.11.0
    > mim install mmcls==0.11.0  # v0.11.0
    # install master branch
    > mim install mmcls -f https://github.com/open-mmlab/mmclassification.git
    # install local repo
    > git clone https://github.com/open-mmlab/mmclassification.git
    > cd mmclassification
    > mim install .

    # install extension based on OpenMMLab
    > mim install mmcls-project -f https://github.com/xxx/mmcls-project.git
    """
    install(package, find_url, timeout, is_yes=is_yes, is_user_dir=is_user_dir)


def install(package: str,
            find_url: str = '',
            timeout: int = 15,
            is_yes: bool = False,
            is_user_dir: bool = False) -> None:
    """Install a package by wheel or from github.

    Args:
        package (str): The name of installed package, such as mmcls.
        find_url (str): Url for finding package. If finding is not provided,
            program will infer the find_url as much as possible. Default: ''.
        timeout (int): The socket timeout. Default: 15.
        is_yes (bool): Don’t ask for confirmation of uninstall deletions.
            Default: False.
        is_usr_dir (bool): Install to the Python user install directory for
            environment variables and user configuration. Default: False.
    """
    target_pkg, target_version = split_package_version(package)

    # mmcv-full -> mmcv
    true_pkg = PKG_ALIAS.get(target_pkg, target_pkg)

    # whether install from local repo
    is_install_local_repo = osp.isdir(osp.abspath(true_pkg)) and not find_url

    # whether install master branch from github
    is_install_master = bool(not target_version and find_url)

    # get target version
    if target_pkg in PKG2PROJECT:
        latest_version = get_latest_version(target_pkg, timeout)
        if target_version:
            if LooseVersion(target_version) > LooseVersion(latest_version):
                error_msg = (f'target_version=={target_version} should not be'
                             f' greater than latest_version=={latest_version}')
                raise ValueError(highlighted_error(error_msg))
        else:
            target_version = latest_version

    # check local environment whether package existed
    if is_install_master or is_install_local_repo:
        pass
    elif is_installed(target_pkg) and target_version:
        existed_version = get_installed_version(target_pkg)
        if is_version_equal(existed_version, target_version):
            echo_warning(f'{target_pkg}=={existed_version} existed.')
            return None
        else:
            if is_yes:
                uninstall(target_pkg, is_yes)
            else:
                confirm_msg = (f'{target_pkg}=={existed_version} has been '
                               f'installed, but want to install {target_pkg}=='
                               f'{target_version}, do you want to uninstall '
                               f'{target_pkg}=={existed_version} and '
                               f'install {target_pkg}=={target_version}? ')
                if click.confirm(confirm_msg):
                    uninstall(target_pkg, True)
                else:
                    echo_warning(f'skip {target_pkg}')
                    return None

    # try to infer find_url if possible
    if not find_url:
        find_url = infer_find_url(target_pkg)

    # whether to write installation records to mmpackage.txt
    is_record = False

    if is_install_local_repo:
        is_record = True
        repo_root = osp.abspath(true_pkg)
        target_pkg, target_version = get_package_version(repo_root)

        echo_success(f'installing {target_pkg} from local repo.')

        if not target_pkg:
            raise FileNotFoundError(
                highlighted_error(f'version.py is missed in {repo_root}'))

        install_from_repo(
            repo_root,
            package=target_pkg,
            timeout=timeout,
            is_yes=is_yes,
            is_user_dir=is_user_dir)

    elif find_url and find_url.find('github.com') >= 0 or is_install_master:
        is_record = True
        install_from_github(target_pkg, target_version, find_url, timeout,
                            is_yes, is_user_dir, is_install_master)
    else:
        # if installing from wheel failed, it will try to install package by
        # building from source if possible.
        is_record = bool(target_pkg in PKG_ALIAS)
        try:
            install_from_wheel(target_pkg, target_version, find_url, timeout,
                               is_user_dir)
        except RuntimeError as error:
            if target_pkg in PKG2PROJECT:
                find_url = f'{DEFAULT_URL}/{PKG2PROJECT[target_pkg]}.git'
                if target_version:
                    target_pkg = f'{target_pkg}=={target_version}'
                if is_yes:
                    install(target_pkg, find_url, timeout, is_yes, is_user_dir)
                else:
                    confirm_msg = (f'install {target_pkg} from wheel, but it '
                                   'failed. Do you want to build it from '
                                   'source if possible?')
                    if click.confirm(confirm_msg):
                        install(target_pkg, find_url, timeout, is_yes,
                                is_user_dir)
                    else:
                        raise RuntimeError(
                            highlighted_error(
                                f'Failed to install {target_pkg}.'))
            else:
                raise RuntimeError(highlighted_error(error))

    if is_record:
        if not target_version:
            target_version = get_installed_version(target_pkg)
        if is_install_local_repo:
            find_url = 'local'

        write_installation_records(target_pkg, target_version, find_url)

    echo_success(f'Successfully installed {target_pkg}.')


def infer_find_url(package: str) -> str:
    """Try to infer find_url if possible.

    If package is the official package, the find_url can be inferred.

    Args:
        package (str): The name of package, such as mmcls.
    """
    find_url = ''
    if package in WHEEL_URL:
        torch_v, cuda_v = get_torch_cuda_version()
        if cuda_v.isdigit():
            cuda_v = f'cu{cuda_v}'
        find_url = WHEEL_URL[package].format(
            cuda_version=cuda_v, torch_version=f'torch{torch_v}')
    elif package in PKG2PROJECT:
        find_url = (f'{DEFAULT_URL}/{PKG2PROJECT[package]}' '.git')

    return find_url


def parse_dependencies(path: str) -> list:
    """Parse dependencies from repo/requirements/mminstall.txt.

    Args:
        path (str): Path of mminstall.txt.
    """

    def _get_proper_version(package, version, op):
        releases = get_release_version(package)
        if op == '>':
            for r_v in releases:
                if LooseVersion(r_v) > LooseVersion(version):
                    return r_v
            else:
                raise ValueError(
                    highlighted_error(f'invalid min version of {package}'))
        elif op == '<':
            for r_v in releases[::-1]:
                if LooseVersion(r_v) < LooseVersion(version):
                    return r_v
            else:
                raise ValueError(
                    highlighted_error(f'invalid max version of {package}'))

    dependencies = []
    with open(path, 'r') as fr:
        for requirement in parse_requirements(fr):
            pkg_name = requirement.project_name
            min_version = ''
            max_version = ''
            for op, version in requirement.specs:
                if op == '==':
                    min_version = max_version = version
                    break
                elif op == '>=':
                    min_version = version
                elif op == '>':
                    min_version = _get_proper_version(pkg_name, version, '>')
                elif op == '<=':
                    max_version = version
                elif op == '<':
                    max_version = _get_proper_version(pkg_name, version, '<')

            dependencies.append([pkg_name, min_version, max_version])

    return dependencies


def install_dependencies(dependencies: List[List[str]],
                         timeout: int = 15,
                         is_yes: bool = False,
                         is_user_dir: bool = False) -> None:
    """Install dependencies, such as mmcls depends on mmcv.

    Args:
        dependencies (list): The list of dependency.
        timeout (int): The socket timeout. Default: 15.
        is_yes (bool): Don’t ask for confirmation of uninstall deletions.
            Default: False.
        is_usr_dir (bool): Install to the Python user install directory for
            environment variables and user configuration. Default: False.
    """
    for target_pkg, min_v, max_v in dependencies:
        target_version = max_v
        latest_version = get_latest_version(target_pkg, timeout)
        if not target_version or LooseVersion(target_version) > LooseVersion(
                latest_version):
            target_version = latest_version

        if is_installed(target_pkg):
            existed_version = get_installed_version(target_pkg)
            if (LooseVersion(min_v) <= LooseVersion(existed_version) <=
                    LooseVersion(target_version)):
                continue

        echo_success(f'installing dependency: {target_pkg}')

        target_pkg = f'{target_pkg}=={target_version}'

        install(
            target_pkg,
            timeout=timeout,
            is_yes=is_yes,
            is_user_dir=is_user_dir)

    echo_success('Successfully installed dependencies.')


def install_from_repo(repo_root: str,
                      *,
                      package: str = '',
                      timeout: int = 15,
                      is_yes: bool = False,
                      is_user_dir: bool = False):
    """Install package from local repo.

    Args:
        repo_root (str): The root of repo.
        package (str): The name of installed package. Default: ''.
        timeout (int): The socket timeout. Default: 15.
        is_yes (bool): Don’t ask for confirmation of uninstall deletions.
            Default: False.
        is_usr_dir (bool): Install to the Python user install directory for
            environment variables and user configuration. Default: False.
    """
    # install dependencies. For example,
    # install mmcls should install mmcv first if it is not installed or
    # its(mmcv) verison does not match.
    mminstall_path = osp.join(repo_root, 'requirements', 'mminstall.txt')
    if osp.exists(mminstall_path):
        dependencies = parse_dependencies(mminstall_path)
        if dependencies:
            install_dependencies(dependencies, timeout, is_yes, is_user_dir)

    true_pkg = PKG_ALIAS.get(package, package)
    pkg_root = osp.join(repo_root, true_pkg)
    src_tool_root = osp.join(repo_root, 'tools')
    dst_tool_root = osp.join(pkg_root, 'tools')
    src_config_root = osp.join(repo_root, 'configs')
    dst_config_root = osp.join(pkg_root, 'configs')
    src_model_zoo_path = osp.join(repo_root, 'model_zoo.yml')
    dst_model_zoo_path = osp.join(pkg_root, 'model_zoo.yml')
    if osp.exists(src_tool_root):
        copy_tree(src_tool_root, dst_tool_root)
    if osp.exists(src_config_root):
        copy_tree(src_config_root, dst_config_root)
    if osp.exists(src_model_zoo_path):
        copy_file(src_model_zoo_path, dst_model_zoo_path)

    third_dependencies = osp.join(repo_root, 'requirements', '/build.txt')
    if osp.exists(third_dependencies):
        dep_cmd = [
            'python', '-m', 'pip', 'install', '-r', third_dependencies,
            '--default-timeout', f'{timeout}'
        ]
        if is_user_dir:
            dep_cmd.append('--user')

        call_command(dep_cmd)

    # write commit id to package
    current_root = os.getcwd()
    os.chdir(repo_root)  # change work dir to repo_root
    commit_id = subprocess.check_output(
        ['git', 'rev-parse', '--short', 'HEAD'])
    os.chdir(current_root)
    commit_id = str(commit_id, 'UTF-8').strip()  # type: ignore
    commit_id_path = os.path.join(pkg_root, 'commit_id.py')
    with open(commit_id_path, 'w') as fw:
        fw.write(f'commit_id = "{commit_id}"')  # type: ignore

    install_cmd = ['python', '-m', 'pip', 'install', repo_root]
    if is_user_dir:
        install_cmd.append('--user')

    # install mmcv with ops by default
    if package in WHEEL_URL or os.getenv('MMCV_WITH_OPS', '1') == 1:
        echo_warning(f'compiling {package} with "MMCV_WITH_OPS=1"')
        os.environ['MMCV_WITH_OPS'] = '1'
        os.environ['MKL_SERVICE_FORCE_INTEL'] = '1'

    call_command(install_cmd)


def install_from_github(package: str,
                        version: str = '',
                        find_url: str = '',
                        timeout: int = 15,
                        is_yes: bool = False,
                        is_user_dir: bool = False,
                        is_install_master: bool = False) -> None:
    """Install package from github.

    Args:
        package (str): The name of installed package, such as mmcls.
        version (str): Version of package. Default: ''.
        find_url (str): Url for finding package. If finding is not provided,
            program will infer the find_url as much as possible. Default: ''.
        timeout (int): The socket timeout. Default: 15.
        is_yes (bool): Don’t ask for confirmation of uninstall deletions.
            Default: False.
        is_usr_dir (bool): Install to the Python user install directory for
            environment variables and user configuration. Default: False.
        is_install_master (bool): Whether install master branch. If it is True,
            process will install master branch. If it is False, process will
            install the specified version. Default: False.
    """
    click.echo(f'installing {package} from {find_url}.')

    _, repo = parse_url(find_url)
    clone_cmd = ['git', 'clone', find_url]
    if not is_install_master:
        clone_cmd.extend(['-b', f'v{version}'])

    with tempfile.TemporaryDirectory() as temp_root:
        repo_root = osp.join(temp_root, repo)
        clone_cmd.append(repo_root)
        call_command(clone_cmd)

        install_from_repo(
            repo_root,
            package=package,
            timeout=timeout,
            is_yes=is_yes,
            is_user_dir=is_user_dir)


def install_from_wheel(package: str,
                       version: str = '',
                       find_url: str = '',
                       timeout: int = 15,
                       is_user_dir: bool = False) -> None:
    """Install wheel from find_url.

    Args:
        package (str): The name of installed package, such as mmcls.
        version (str): Version of package. Default: ''.
        find_url (str): Url for finding package. If finding is not provided,
            program will infer the find_url as much as possible. Default: ''.
        timeout (int): The socket timeout. Default: 15.
        is_usr_dir (bool): Install to the Python user install directory for
            environment variables and user configuration. Default: False.
    """
    click.echo(f'installing {package} from wheel.')

    install_cmd = [
        'python', '-m', 'pip', '--default-timeout', f'{timeout}', 'install'
    ]
    if version:
        install_cmd.append(f'{package}=={version}')
    else:
        install_cmd.append(package)
    if find_url:
        install_cmd.extend(['-f', find_url])
    if is_user_dir:
        install_cmd.append('--user')

    call_command(install_cmd)



import difflib
import os
import os.path as osp

import click

plugin_folder = os.path.join(os.path.dirname(__file__), 'commands')
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class MIM(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if not filename.startswith('__') and filename.endswith('.py'):
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = osp.join(plugin_folder, name + '.py')
        if not osp.exists(fn):
            matches = [
                x for x in self.list_commands(ctx) if x.startswith(name)
            ]
            if not matches:
                return None
            elif len(matches) == 1:
                return self.get_command(ctx, matches[0])

        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            eval(code, ns, ns)
        return ns['cli']

    def resolve_command(self, ctx, args):
        # The implementation is modified from https://github.com/click-contrib/
        # click-didyoumean/blob/master/click_didyoumean/__init__.py#L25
        try:
            return super().resolve_command(ctx, args)
        except click.exceptions.UsageError as error:
            error_msg = str(error)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(original_cmd_name,
                                                self.list_commands(ctx), 3,
                                                0.1)
            if matches:
                error_msg += '\n\nDid you mean one of these?\n'
                error_msg += '\n'.join(matches)

            raise click.exceptions.UsageError(error_msg, error.ctx)


@click.command(cls=MIM, context_settings=CONTEXT_SETTINGS)
@click.version_option()
def cli():
    """OpenMMLab Command Line Interface.

    MIM provides a unified API for launching and installing OpenMMLab projects
    and their extensions, and managing the OpenMMLab model zoo.
    """
    pass



import os
import site

USER = 'open-mmlab'
DEFAULT_URL = f'https://github.com/{USER}'

WHEEL_URL = {
    'mmcv-full':
    'https://download.openmmlab.com/mmcv/dist/{cuda_version}/'
    '{torch_version}/index.html',
}
RAW_GITHUB_URL = 'https://raw.githubusercontent.com/{owner}/{repo}/{branch}'
PKG2PROJECT = {
    'mmcv-full': 'mmcv',
    'mmcls': 'mmclassification',
    'mmdet': 'mmdetection',
    'mmdet3d': 'mmdetection3d',
    'mmsegmentation': 'mmsegmentation',
    'mmaction2': 'mmaction2',
    'mmtrack': 'mmtracking',
    'mmpose': 'mmpose',
    'mmedit': 'mmediting',
    'mmocr': 'mmocr',
    'mmgen': 'mmgeneration',
}
PKG_ALIAS = {
    'mmcv-full': 'mmcv',
    'mmcv': 'mmcv',
}

DEFAULT_CACHE_DIR = f'{os.environ["HOME"]}/.cache/mim'
if not os.path.exists(DEFAULT_CACHE_DIR):
    os.makedirs(DEFAULT_CACHE_DIR)

MMPACKAGE_PATH = os.path.join(site.getsitepackages()[0], 'mmpackage.txt')



import os
import os.path as osp
import random as rd
import subprocess
from typing import Optional, Tuple, Union

import click

from mim.click import CustomCommand, param2lowercase
from mim.utils import (
    echo_success,
    exit_with_error,
    get_installed_path,
    highlighted_error,
    is_installed,
    recursively_find,
)


@click.command(
    'test',
    context_settings=dict(ignore_unknown_options=True),
    cls=CustomCommand)
@click.argument('package', type=str, callback=param2lowercase)
@click.argument('config', type=str)
@click.option('--checkpoint', type=str, default=None, help='checkpoint path')
@click.option(
    '--launcher',
    type=click.Choice(['none', 'pytorch', 'slurm'], case_sensitive=False),
    default='none',
    help='Job launcher')
@click.option(
    '--port',
    type=int,
    default=None,
    help=('The port used for inter-process communication (only applicable to '
          'slurm / pytorch launchers). If set to None, will randomly choose '
          'a port between 20000 and 30000'))
@click.option(
    '--gpus',
    type=int,
    help='Number of gpus to use (only applicable to launcher == "slurm")')
@click.option(
    '--gpus-per-node',
    type=int,
    help=('Number of gpus per node to use '
          '(only applicable to launcher == "slurm")'))
@click.option(
    '--cpus-per-task',
    type=int,
    default=2,
    help='Number of cpus per task (only applicable to launcher == "slurm")')
@click.option(
    '--partition',
    type=str,
    help='The partition to use (only applicable to launcher == "slurm")')
@click.option(
    '--srun-args', type=str, help='Other srun arguments that might be used')
@click.argument('other_args', nargs=-1, type=click.UNPROCESSED)
def cli(package: str,
        config: str,
        checkpoint: str,
        gpus: int,
        gpus_per_node: int,
        partition: str,
        cpus_per_task: int = 2,
        launcher: str = 'none',
        port: int = None,
        srun_args: Optional[str] = None,
        other_args: tuple = ()) -> None:
    """Perform Testing.

    Example:

    \b
    # Test models on a single server with 1 GPU, report accuracy
    > mim test mmcls resnet101_b16x8_cifar10.py --checkpoint \
        tmp/epoch_3.pth --gpus 1 --metrics accuracy
    # Test models on a single server with 1 GPU, save predictions
    > mim test mmcls resnet101_b16x8_cifar10.py --checkpoint \
        tmp/epoch_3.pth --gpus 1 --out tmp.pkl
    # Test models on a single server with 4 GPUs, pytorch distributed,
    # report accuracy
    > mim test mmcls resnet101_b16x8_cifar10.py --checkpoint \
        tmp/epoch_3.pth --gpus 4 --launcher pytorch --metrics accuracy
    # Test models on a slurm HPC with one 8-GPU node, report accuracy
    > mim test mmcls resnet101_b16x8_cifar10.py --checkpoint \
        tmp/epoch_3.pth --gpus 8 --metrics accuracy --partition \
        partition_name --gpus-per-node 8 --launcher slurm
    # Print help messages of sub-command test
    > mim test -h
    # Print help messages of sub-command test and the testing script of mmcls
    > mim test mmcls -h
    """

    is_success, msg = test(
        package=package,
        config=config,
        checkpoint=checkpoint,
        gpus=gpus,
        gpus_per_node=gpus_per_node,
        cpus_per_task=cpus_per_task,
        partition=partition,
        launcher=launcher,
        port=port,
        srun_args=srun_args,
        other_args=other_args)

    if is_success:
        echo_success(msg)  # type: ignore
    else:
        exit_with_error(msg)


def test(
    package: str,
    config: str,
    checkpoint: str = None,
    gpus: int = None,
    gpus_per_node: int = None,
    cpus_per_task: int = 2,
    partition: str = None,
    launcher: str = 'none',
    port: int = None,
    srun_args: Optional[str] = None,
    other_args: tuple = ()
) -> Tuple[bool, Union[str, Exception]]:
    """Test a model with given config.

    Args:
        package (str): The codebase name.
        config (str): The config file path. If not exists, will search in the
            config files of the codebase.
        checkpoint (str): The path to the checkpoint file. Default to None.
        gpus (int): Number of gpus used for testing
            (only applicable to launcher == "slurm"). Defaults to None.
        gpus_per_node (int, optional): Number of gpus per node to use
            (only applicable to launcher == "slurm"). Defaults to None.
        cpus_per_task (int, optional): Number of cpus per task to use
            (only applicable to launcher == "slurm"). Defaults to None.
        partition (str, optional): The partition name
            (only applicable to launcher == "slurm"). Defaults to None.
        launcher (str, optional): The launcher used to launch jobs.
            Defaults to 'none'.
        port (int | None, optional): The port used for inter-process
            communication (only applicable to slurm / pytorch launchers).
            Default to None. If set to None, will randomly choose a port
            between 20000 and 30000.
        srun_args (str, optional): Other srun arguments that might be
            used, all arguments should be in a string. Defaults to None.
        other_args (tuple, optional): Other arguments, will be passed to the
            codebase's training script. Defaults to ().
    """

    # `checkpoint` is a compulsory argument for all mm codebases except
    # mmtracking, so that if the codebase is not mmtracking, user must specify
    # the checkpoint.
    # [TODO]: refactor the code logic and remove the hard coding
    if checkpoint is None:
        assert package == 'mmtrack', (
            'You must specify the checkpoint path '
            'unless you are testing mmtracking models')

    if port is None:
        port = rd.randint(20000, 30000)

    if launcher in ['slurm', 'pytorch']:
        click.echo(f'Using port {port} for synchronization. ')

    # If launcher == "slurm", must have following args
    if launcher == 'slurm':
        msg = ('If launcher is slurm, '
               'gpus, gpus-per-node and partition should not be None')
        flag = (gpus_per_node
                is not None) and (partition is not None) and (gpus is not None)
        assert flag, msg

    if not is_installed(package):
        msg = (f'The codebase {package} is not installed, '
               'do you want to install it? ')
        if click.confirm(msg):
            click.echo(f'Installing {package}')
            cmd = ['mim', 'install', package]
            ret = subprocess.check_call(cmd)
            if ret != 0:
                msg = f'{package} is not successfully installed'
                raise RuntimeError(highlighted_error(msg))
            else:
                click.echo(f'{package} is successfully installed')
        else:
            msg = f'You can not test this model without {package} installed.'
            return False, msg

    pkg_root = get_installed_path(package)

    if not osp.exists(config):
        files = recursively_find(pkg_root, osp.basename(config))

        if len(files) == 0:
            msg = (f"The path {config} doesn't exist and we can not find "
                   f'the config file in codebase {package}.')
            raise ValueError(highlighted_error(msg))
        elif len(files) > 1:
            msg = (
                f"The path {config} doesn't exist and we find multiple "
                f'config files with same name in codebase {package}: {files}.')
            raise ValueError(highlighted_error(msg))
        click.echo(
            f"The path {config} doesn't exist but we find the config file "
            f'in codebase {package}, will use {files[0]} instead.')
        config = files[0]

    # We know that 'config' exists and is legal.
    test_script = osp.join(pkg_root, 'tools/test.py')

    common_args = ['--launcher', launcher] + list(other_args)

    if launcher == 'none':
        cmd = ['python', test_script, config]
        if checkpoint:
            cmd += [checkpoint]
        cmd += common_args
    elif launcher == 'pytorch':
        cmd = [
            'python', '-m', 'torch.distributed.launch',
            f'--nproc_per_node={gpus}', f'--master_port={port}', test_script,
            config
        ]
        if checkpoint:
            cmd += [checkpoint]
        cmd += common_args
    elif launcher == 'slurm':
        parsed_srun_args = srun_args.split() if srun_args else []
        cmd = [
            'srun', '-p', f'{partition}', f'--gres=gpu:{gpus_per_node}',
            f'--ntasks={gpus}', f'--ntasks-per-node={gpus_per_node}',
            f'--cpus-per-task={cpus_per_task}', '--kill-on-bad-exit=1'
        ]
        cmd += parsed_srun_args
        cmd += ['python', '-u', test_script, config]
        if checkpoint:
            cmd += [checkpoint]
        cmd += common_args

    cmd_text = ' '.join(cmd)
    click.echo(f'Testing command is {cmd_text}. ')
    ret = subprocess.check_call(
        cmd, env=dict(os.environ, MASTER_PORT=str(port)))
    if ret == 0:
        return True, 'Testing finished successfully.'
    else:
        return False, 'Testing not finished successfully.'



from .default import (
    DEFAULT_CACHE_DIR,
    DEFAULT_URL,
    PKG2PROJECT,
    PKG_ALIAS,
    RAW_GITHUB_URL,
    USER,
    WHEEL_URL,
)
from .utils import (
    MMPACKAGE_PATH,
    args2string,
    call_command,
    cast2lowercase,
    color_echo,
    download_from_file,
    echo_error,
    echo_success,
    echo_warning,
    exit_with_error,
    extract_tar,
    get_commit_id,
    get_config,
    get_content_from_url,
    get_github_url,
    get_installed_path,
    get_installed_version,
    get_latest_version,
    get_package_version,
    get_release_version,
    get_torch_cuda_version,
    highlighted_error,
    is_installed,
    is_version_equal,
    parse_url,
    read_installation_records,
    recursively_find,
    remove_installation_records,
    set_config,
    split_package_version,
    string2args,
    write_installation_records,
)

__all__ = [
    'DEFAULT_CACHE_DIR',
    'DEFAULT_URL',
    'cast2lowercase',
    'echo_error',
    'echo_warning',
    'echo_success',
    'exit_with_error',
    'get_content_from_url',
    'get_github_url',
    'get_installed_version',
    'get_installed_path',
    'get_latest_version',
    'get_torch_cuda_version',
    'is_installed',
    'parse_url',
    'PKG2PROJECT',
    'PKG_ALIAS',
    'RAW_GITHUB_URL',
    'read_installation_records',
    'write_installation_records',
    'remove_installation_records',
    'recursively_find',
    'color_echo',
    'USER',
    'WHEEL_URL',
    'split_package_version',
    'call_command',
    'get_commit_id',
    'is_version_equal',
    'MMPACKAGE_PATH',
    'get_package_version',
    'string2args',
    'args2string',
    'get_config',
    'set_config',
    'download_from_file',
    'highlighted_error',
    'extract_tar',
    'get_release_version',
]



from setuptools import find_packages, setup  # type: ignore

setup(
    name='openmim',
    version='0.1.0',
    description='MIM Install OpenMMLab packages',
    author='OpenMMLab',
    author_email='openmmlab@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'Click==7.1.2',
        'colorama',
        'requests',
        'model-index',
        'pandas',
        'tabulate',
    ],
    entry_points='''
        [console_scripts]
        mim=mim.cli:cli
    ''',
)



import os.path as osp
from pkg_resources import resource_filename
from typing import List, Optional

import click

from mim.click import OptionEatAll, get_downstream_package, param2lowercase
from mim.commands.search import get_model_info
from mim.utils import (
    DEFAULT_CACHE_DIR,
    download_from_file,
    echo_success,
    highlighted_error,
    is_installed,
    split_package_version,
)


@click.command('download')
@click.argument(
    'package',
    type=str,
    autocompletion=get_downstream_package,
    callback=param2lowercase)
@click.option(
    '--config',
    'configs',
    cls=OptionEatAll,
    required=True,
    help='Config ids to download, such as resnet18_b16x8_cifar10')
@click.option(
    '--dest', 'dest_root', type=str, help='Destination of saving checkpoints.')
def cli(package: str,
        configs: List[str],
        dest_root: Optional[str] = None) -> None:
    """Download checkpoints from url and parse configs from package.

    \b
    Example:
        > mim download mmcls --config resnet18_b16x8_cifar10
        > mim download mmcls --config resnet18_b16x8_cifar10 --dest .
    """
    download(package, configs, dest_root)


def download(package: str,
             configs: List[str],
             dest_root: Optional[str] = None) -> List[str]:
    """Download checkpoints from url and parse configs from package.

    Args:
        package (str): Name of package.
        configs (List[str]): List of config ids.
        dest_root (Optional[str]): Destination directory to save checkpoint and
            config. Default: None.
    """
    if dest_root is None:
        dest_root = DEFAULT_CACHE_DIR

    dest_root = osp.abspath(dest_root)

    package, version = split_package_version(package)
    if version:
        raise ValueError('version is not allowed, please type '
                         '"mim download -h" to show the correct way.')

    if not is_installed(package):
        raise RuntimeError(f'{package} is not installed. Please install it '
                           'first.')

    checkpoints = []
    model_info = get_model_info(
        package, shown_fields=['weight', 'config'], to_dict=True)
    valid_configs = model_info.keys()
    invalid_configs = set(configs) - set(valid_configs)
    if invalid_configs:
        raise ValueError(f'Expected configs: {valid_configs}, but got '
                         f'{invalid_configs}')

    from mmcv import Config

    for config in configs:
        click.echo(f'processing {config}...')

        checkpoint_urls = model_info[config]['weight']
        for checkpoint_url in checkpoint_urls.split(','):
            filename = checkpoint_url.split('/')[-1]
            checkpoint_path = osp.join(dest_root, filename)
            if osp.exists(checkpoint_path):
                echo_success(f'{filename} exists in {dest_root}')
            else:
                # TODO: check checkpoint hash when all the models are ready.
                download_from_file(checkpoint_url, checkpoint_path)

                echo_success(
                    f'Successfully downloaded {filename} to {dest_root}')

        config_paths = model_info[config]['config']
        for config_path in config_paths.split(','):
            config_path = resource_filename(package, config_path)
            if not osp.exists(config_path):
                raise ValueError(
                    highlighted_error(f'{config_path} is not found.'))

            config_obj = Config.fromfile(config_path)
            saved_config_path = osp.join(dest_root, f'{config}.py')
            config_obj.dump(saved_config_path)
            echo_success(f'Successfully dumped {config}.py to {dest_root}')

            checkpoints.append(filename)

    return checkpoints



# flake8: noqa
import os
import os.path as osp
import subprocess
from typing import Tuple, Union

import click

from mim.click import CustomCommand, param2lowercase
from mim.utils import (
    echo_success,
    echo_warning,
    exit_with_error,
    get_installed_path,
    highlighted_error,
    is_installed,
    recursively_find,
)


@click.command(
    'run',
    context_settings=dict(ignore_unknown_options=True),
    cls=CustomCommand)
@click.argument('package', type=str, callback=param2lowercase)
@click.argument('command', type=str)
@click.argument('other_args', nargs=-1, type=click.UNPROCESSED)
def cli(package: str, command: str, other_args: tuple = ()) -> None:
    """Run arbitrary command of a codebase.

    Example:

    \b
    # Get the Flops of a model
    > mim run mmcls get_flops resnet101_b16x8_cifar10.py
    # Publish a model
    > mim run mmcls publish_model input.pth output.pth
    # Train models on a slurm HPC with one GPU
    > srun -p partition --gres=gpu:1 mim run mmcls train \
        resnet101_b16x8_cifar10.py --work-dir tmp
    # Test models on a slurm HPC with one GPU, report accuracy
    > srun -p partition --gres=gpu:1 mim run mmcls test \
        resnet101_b16x8_cifar10.py tmp/epoch_3.pth --metrics accuracy
    # Print help messages of sub-command run
    > mim run -h
    # Print help messages of sub-command run, list all available scripts in
    # codebase mmcls
    > mim run mmcls -h
    # Print help messages of sub-command run, print the help message of
    # training script in mmcls
    > mim run mmcls train -h
    """
    is_success, msg = run(
        package=package, command=command, other_args=other_args)

    if is_success:
        echo_success(msg)  # type: ignore
    else:
        exit_with_error(msg)


def run(
    package: str, command: str,
    other_args: tuple = ()) -> Tuple[bool, Union[str, Exception]]:
    """Run arbitrary command of a codebase.

    This command assumes the command scripts have been put into the
    ``package/tools`` directory.

    Args:
        package (str): The codebase name.
        other_args (tuple, optional): Other arguments, will be passed to the
            codebase's script. Defaults to ().
    """

    if not is_installed(package):
        msg = (f'The codebase {package} is not installed, '
               'do you want to install it? ')
        if click.confirm(msg):
            click.echo(f'Installing {package}')
            cmd = ['mim', 'install', package]
            ret = subprocess.check_call(cmd)
            if ret != 0:
                msg = f'{package} is not successfully installed'
                raise RuntimeError(highlighted_error(msg))
            else:
                click.echo(f'{package} is successfully installed')
        else:
            msg = (f"You can't run commands in {package} without "
                   f'{package} installed.')
            return False, msg

    pkg_root = get_installed_path(package)

    prefix = osp.join(pkg_root, 'tools/')

    command_domain = ''
    if ':' in command:
        split_command = command.split(':')
        command_domain = '/'.join(split_command[:-1])
        command = split_command[-1]

    files = recursively_find(prefix, command + '.py')

    if command_domain == '':
        suffix = f'/{command}.py'
    else:
        suffix = f'/{command_domain}/{command}.py'
    files = [f for f in files if f.endswith(suffix)]

    if len(files) == 0:
        msg = f'No script in codebase {package} has suffix {suffix}.'
        raise ValueError(highlighted_error(msg))
    elif len(files) > 1:
        echo_warning(
            f'Multiple scripts in codebase {package} have suffix {suffix}: ')
        for f in files:
            echo_warning(f)

        # Use the shortest path
        files.sort(key=lambda x: len(x.split('/')))
        echo_warning(f'We are using the script {files[0]}. ')
        echo_warning('To use other scripts, you need to use these commands: ')
        for f in files[1:]:
            cmd = f.split(prefix)[1].split('.')[0].replace('/', ':')
            echo_warning(f'Command for {f}: {cmd}')

    script = files[0]
    click.echo(f'Use the script {script} for command {command}.')

    cmd = ['python', script] + list(other_args)

    cmd_text = ' '.join(cmd)
    click.echo(f'The command to call is {cmd_text}. ')
    ret = subprocess.check_call(cmd, env=dict(os.environ))
    if ret == 0:
        return True, 'The script finished successfully.'
    else:
        return False, 'The script not finished successfully.'
